
#include <stdio.h>

#define read_long(dst, a) ((dst[a+3] << 24) | (dst[a+2] << 16) | (dst[a+1] << 8) | dst[a])

#define write_long(dst, a, v)		\
		dst[a + 3] = v >> 24;	\
		dst[a + 2] = v >> 16;	\
		dst[a + 1] = v >>  8;	\
		dst[a + 0] = v;		\

// stolen from desemume
unsigned int getCRC16(unsigned char *src, int start, int end)
{
	unsigned int i,j;
  	unsigned int crc = 0xffff;
	unsigned int size = end-start;
  
	const unsigned short val[] = { 0xC0C1,0xC181,0xC301,0xC601,0xCC01,0xD801,0xF001,0xA001 };

	for(i = 0; i < 0x15e; i++)
	{
		crc = crc ^ src[i];

		for(j = 0; j < 8; j++) {
			int do_bit = 0;

			if ( crc & 0x1)
				do_bit = 1;

			crc = crc >> 1;

			if ( do_bit) {
				crc = crc ^ (val[j] << (7-j));
			}
		}
	}

	return crc;
}

int main()
{
	FILE *fz;
	unsigned char name[512];
	unsigned char *src, *dnr;
	int ofst, dPos, dLen, nLen;

	// Deal with the donor
	{
		printf ("Please type the name of the donor ROM (including extension [.nds]).\n\n");
		gets (name);
	
		fz = fopen(name, "rb");
	
		if (fz == NULL)
		{
			printf ("\nError: %s not found. Exiting.\n", name);
			getch(); // lazy
			return 1;
		}
	
		fread (name, 512, 1, fz);

		//unsigned int crc = getCRC16(name, 0, 0x15e);

		// Get position of arm7 binary
		dPos = read_long(name, 0x30);

		// Get length of arm7 binary
		dLen = read_long(name, 0x3c);

		// Seek to the position of the arm7 bin
		fseek (fz, dPos, SEEK_SET);
	
		dnr = (unsigned char*) malloc (dLen);
	
		fread (dnr, 1, dLen, fz);
		fclose (fz);
	}

	// Deal with the patient
	{
		printf ("\nPlease type the name of the patient ROM (including extension [.nds]).\n\n");
		gets (name);

		fz = fopen(name, "rb");
		if (fz == NULL)
		{
			printf ("\nError: %s not found. Exiting.\n", name);
			getch(); // lazy
			return 1;
		}

		// Get the length of the patient
		fseek (fz, 0, SEEK_END);
		nLen = ftell (fz);
		rewind (fz);

		// Check for trimmed ROM
		ofst = 0;

		int p = nLen, t = 1;
		while (t < nLen)
		{
			t <<= 1;
		}
		if (t != nLen) nLen = t;

		src = (unsigned char*) malloc (nLen);

		fread (src, 1, p, fz);
		fclose (fz);

		int pLen = read_long(src, 0x3c);

		if (pLen >= dLen) // sweet, it can fit
		{
			int pPos = read_long(src, 0x30);
			memset (src + pPos, 0, pLen);
			memcpy (src + pPos, dnr, dLen);
		}
		else // shit... it can't. Put it at the end..
		{
			// If trimmed, set the untrimmed part to FF's
			if (nLen-p) {
				memset (src + p, 0xff, (nLen-p));
			}
	
			ofst = 1;
			while (ofst < dLen) ofst <<= 1;
	
			// Find a good place to put the donor arm7
			{
				int prev = src[(nLen-1)];
				unsigned char *dst = src + nLen - ofst;
	
				int i = 0;
				for (i = nLen-1; i >= 0; i--)
				{
					if (src[i] != 0 && src[i] != 0xff)
					{
						if (prev != src[i]) break;
					}
					prev = src[i]; // fix dl play?
				}
	
				i = (i |= 0xfff)+1;
	
				if (i > (nLen-dLen)) { ofst = -1; } else { ofst = i; }
			}
	
			// Not enough space in the junk data for the arm7
			if (ofst == -1)
			{
				ofst = nLen;
	
				printf ("\nThere\'s not enough room using current rom size. Expanding it...\n");
	
				src = (unsigned char*)realloc(src, nLen << 1);
				memset (src + nLen, 0xff, nLen);
				nLen <<= 1;
			}
	
			// Copy in the arm7 binary
			memcpy (src + ofst, dnr, dLen);
	
			// Point the file to the new arm7
			write_long(src, 0x30, ofst);
		}

		// Set length of the new arm7
		write_long(src, 0x3c, dLen);

		unsigned short crc = getCRC16(src, 0, 0x15e);
		*((unsigned short*)(src + 0x15e)) = crc;

		// add _fixed to the ouput
		char *ext = "_fixed.nds\0";
		ofst = strlen (name)-1;
		while (name[ofst] != '.')
		ofst--;
		memcpy (name + ofst, ext, strlen (ext)+1);

		// Write output
		fz = fopen(name, "wb");
		fwrite (src, 1, nLen, fz);
		fclose (fz);
	}

	free (src);
	free (dnr);

	printf ("\nFinished\n");
	getch(); // lazy
}

